@@ -83,6 +83,8 @@ def _mv_legacy_headers_for_service(mv_service_type):
         headers.append("X-OpenStack-Nova-API-Version")
     elif mv_service_type == "baremetal":
         headers.append("X-OpenStack-Ironic-API-Version")
+    elif mv_service_type in ["sharev2", "shared-file-system"]:
+        headers.append("X-OpenStack-Manila-API-Version")
     return headers
 
 
@@ -330,6 +332,9 @@ class Session(object):
     :param rate_semaphore: Semaphore to be used to control concurrency
                            and rate limiting of requests. (optional,
                            defaults to no concurrency or rate control)
+    :param int connect_retries: the maximum number of retries that should
+                                be attempted for connection errors.
+                                (optional, defaults to 0 - never retry).
     """
 
     user_agent = None
@@ -343,10 +348,18 @@ class Session(object):
                  redirect=_DEFAULT_REDIRECT_LIMIT, additional_headers=None,
                  app_name=None, app_version=None, additional_user_agent=None,
                  discovery_cache=None, split_loggers=None,
-                 collect_timing=False, rate_semaphore=None):
+                 collect_timing=False, rate_semaphore=None,
+                 connect_retries=0):
 
         self.auth = auth
         self.session = _construct_session(session)
+        # NOTE(mwhahaha): keep a reference to the session object so we can
+        # clean it up when this object goes away. We don't want to close the
+        # session if it was passed into us as it may be reused externally.
+        # See LP#1838704
+        self._session = None
+        if not session:
+            self._session = self.session
         self.original_ip = original_ip
         self.verify = verify
         self.cert = cert
@@ -364,6 +377,7 @@ class Session(object):
         # so we can distinguish between the value being set or not.
         self._split_loggers = split_loggers
         self._collect_timing = collect_timing
+        self._connect_retries = connect_retries
         self._api_times = []
         self._rate_semaphore = rate_semaphore or NoOpSemaphore()
 
@@ -375,6 +389,17 @@ class Session(object):
 
         self._json = _JSONEncoder()
 
+    def __del__(self):
+        """Clean up resources on delete."""
+        if self._session:
+            # If we created a requests.Session, try to close it out correctly
+            try:
+                self._session.close()
+            except Exception:
+                pass
+            finally:
+                self._session = None
+
     @property
     def adapters(self):
         return self.session.adapters
@@ -591,6 +616,14 @@ class Session(object):
         if (service_type.startswith('volume') or
                 service_type == 'block-storage'):
             service_type = 'volume'
+        elif service_type.startswith('share'):
+            # NOTE(gouthamr) manila doesn't honor the "OpenStack-API-Version"
+            # header yet, but sending it does no harm - when the service
+            # honors this header, it'll use the standardized name in the
+            # service-types-authority and not the legacy name in the cloud's
+            # service catalog
+            service_type = 'shared-file-system'
+
         headers.setdefault('OpenStack-API-Version',
                            '{service_type} {microversion}'.format(
                                service_type=service_type,
@@ -603,7 +636,7 @@ class Session(object):
                 user_agent=None, redirect=None, authenticated=None,
                 endpoint_filter=None, auth=None, requests_auth=None,
                 raise_exc=True, allow_reauth=True, log=True,
-                endpoint_override=None, connect_retries=0, logger=None,
+                endpoint_override=None, connect_retries=None, logger=None,
                 allow=None, client_name=None, client_version=None,
                 microversion=None, microversion_service_type=None,
                 status_code_retries=0, retriable_status_codes=None,
@@ -636,7 +669,7 @@ class Session(object):
                                   for forever/never. (optional)
         :param int connect_retries: the maximum number of retries that should
                                     be attempted for connection errors.
-                                    (optional, defaults to 0 - never retry).
+                                    (optional, defaults to None - never retry).
         :param bool authenticated: True if a token should be attached to this
                                    request, False if not or None for attach if
                                    an auth_plugin is available.
@@ -730,6 +763,16 @@ class Session(object):
         else:
             split_loggers = None
         logger = logger or utils.get_logger(__name__)
+        # NOTE(gmann): Convert r initlize the headers to
+        # CaseInsensitiveDict to make sure headers are
+        # case insensitive.
+        if kwargs.get('headers'):
+            kwargs['headers'] = requests.structures.CaseInsensitiveDict(
+                kwargs['headers'])
+        else:
+            kwargs['headers'] = requests.structures.CaseInsensitiveDict()
+        if connect_retries is None:
+            connect_retries = self._connect_retries
         # HTTP 503 - Service Unavailable
         retriable_status_codes = retriable_status_codes or [503]
         rate_semaphore = rate_semaphore or self._rate_semaphore
